@@ -1,11 +1,23 @@
 """The core library functionality."""
+from __future__ import annotations
 
 from collections.abc import Iterable
 from copy import deepcopy
 from dataclasses import dataclass
 from datetime import timedelta
+from functools import partial
 from itertools import product
-from typing import Any, Callable, Dict, Generic, List, Optional, Sequence, TypeVar
+from typing import (
+    Any,
+    Callable,
+    Dict,
+    Generic,
+    List,
+    Optional,
+    Sequence,
+    TypeVar,
+    Tuple,
+)
 from unittest import TestCase, TestSuite
 from unittest.mock import patch
 
@@ -26,14 +38,31 @@ AGA_RESERVED_KEYWORDS = {
     "aga_extra_credit",
     "aga_override_check",
     "aga_override_test",
+    "aga_description",
 }
 
 
+__all__ = [
+    "TestMetadata",
+    "SubmissionMetadata",
+    "AgaTestCase",
+    "AgaTestSuite",
+    "Problem",
+    "problem",
+    "test_case",
+    "param",
+    "test_cases",
+    "test_cases_params",
+    "test_cases_zip",
+    "test_cases_product",
+    "group",
+]
+
+
 @dataclass(frozen=True)
 class TestMetadata:
     """Stores metadata about a specific test case."""
 
-    name: str
     max_score: float
     config: AgaTestConfig
     check_stdout: bool
@@ -82,10 +111,16 @@ class AgaTestCase(TestCase):
         self._under_test = under_test
         self._metadata = metadata
 
+    @property
     def metadata(self) -> TestMetadata:
         """Get the test's metadata."""
         return self._metadata
 
+    @property
+    def test_input(self) -> _TestInputs[Output]:
+        """Get the test input."""
+        return self._test_input
+
     # pylint: disable=invalid-name
     # camelCase is required by unittest
     def runTest(self) -> None:
@@ -99,7 +134,21 @@ class AgaTestCase(TestCase):
 
         This method is called by unittest.
         """
-        return self._metadata.name
+        return self.name
+
+    @property
+    def name(self) -> str:
+        """Format the name of the test case."""
+        return self._test_input.name or self._metadata.config.name_fmt.format(
+            args=self._test_input.args_repr(self._metadata.config.name_sep),
+            kwargs=self._test_input.kwargs_repr(self._metadata.config.name_sep),
+            sep=self._test_input.sep(self._metadata.config.name_sep),
+        )
+
+    @property
+    def description(self) -> str | None:
+        """Get the problem's description."""
+        return self._test_input.description
 
 
 class AgaTestSuite(TestSuite):
@@ -147,6 +196,7 @@ class _TestInputs(TestCase, Generic[Output]):
         aga_override_test: Optional[
             Callable[[TestCase, Callable[..., Output], Callable[..., Output]], None]
         ],
+        aga_description: Optional[str],
         **kwargs: Any,
     ) -> None:
         super().__init__()
@@ -157,11 +207,42 @@ class _TestInputs(TestCase, Generic[Output]):
         self._expect_stdout = aga_expect_stdout
         self._override_check = aga_override_check
         self._override_test = aga_override_test
+        self._description = aga_description
         self.score_info = ScoreInfo(aga_weight, aga_value, aga_extra_credit)
 
         self._args = args
         self._kwargs = kwargs
 
+    @property
+    def args(self) -> Tuple[Any, ...]:
+        """Get the positional arguments for the test case."""
+        return self._args
+
+    @property
+    def kwargs(self) -> Dict[str, Any]:
+        """Get the keyword arguments for the test case."""
+        return self._kwargs
+
+    @property
+    def description(self) -> str | None:
+        """Get the description of the test case."""
+        return self._description
+
+    @description.setter
+    def description(self, desc: str | None) -> None:
+        """Set the description of the test case."""
+        self._description = desc
+
+    @property
+    def name(self) -> str | None:
+        """Get the name of the test case."""
+        return self._name
+
+    @name.setter
+    def name(self, name: str | None) -> None:
+        """Set the name of the test case."""
+        self._name = name
+
     def _eval(self, func: Callable[..., Any]) -> Any:
         """Evaluate func on the arguments."""
         # deepcopy in case the student submission mutates arguments; we don't want it to
@@ -253,16 +334,7 @@ class _TestInputs(TestCase, Generic[Output]):
         config: AgaConfig,
     ) -> AgaTestCase:
         """Generate a TestCase which tests `golden` against `under_test`."""
-        name_fmt = config.test.name_fmt
-        name_sep = config.test.name_sep
-
-        name = self._name or name_fmt.format(
-            args=self._args_repr(name_sep),
-            kwargs=self._kwargs_repr(name_sep),
-            sep=self._sep(name_sep),
-        )
         metadata = TestMetadata(
-            name=name,
             hidden=self._hidden,
             config=config.test,
             max_score=score,
@@ -271,41 +343,58 @@ class _TestInputs(TestCase, Generic[Output]):
         )
         return AgaTestCase(self, golden, under_test, metadata)
 
-    def _args_repr(self, sep: str) -> str:
+    def args_repr(self, sep: str) -> str:
+        """Get a string representation of the arguments."""
         return sep.join(repr(x) for x in self._args)
 
-    def _kwargs_repr(self, sep: str) -> str:
+    def kwargs_repr(self, sep: str) -> str:
+        """Get a string representation of the keyword arguments."""
         # we use k instead of repr(k) so we don't get quotes around it
         return sep.join(k + "=" + repr(v) for k, v in self._kwargs.items())
 
-    def _sep(self, sep: str) -> str:
+    def sep(self, sep: str) -> str:
         """Return sep if both exist, "" otherwise."""
         assert sep == ","
         return self._args and self._kwargs and sep or ""
 
     def __repr__(self) -> str:
-        args_repr = self._args_repr(",")
-        kwargs_repr = self._kwargs_repr(",")
-        sep = self._sep(",")
+        args_repr = self.args_repr(",")
+        kwargs_repr = self.kwargs_repr(",")
+        sep = self.sep(",")
 
         return args_repr + sep + kwargs_repr
 
     def check_one(self, golden: Callable[..., Output]) -> None:
         """Check that the golden solution is correct."""
         if self._expect is not None or self._expect_stdout is not None:
-            golden_stdout, golden_output = self._eval(
-                with_captured_stdout(golden)
-            )  # type: str, Output
-            if self._expect is not None:
-                self.assertEqual(golden_output, self._expect)
+            if self._override_test:
 
-            if self._expect_stdout is not None:
-                if isinstance(self._expect_stdout, str):
-                    self.assertEqual(golden_stdout, self._expect_stdout)
-                elif isinstance(self._expect_stdout, Sequence):
-                    self.assertEqual(
-                        golden_stdout.splitlines(), list(self._expect_stdout)
-                    )
+                def dummy_tester(*_: Any, **__: Any) -> Output:
+                    # https://github.com/python/mypy/issues/4805 ehh
+                    return self._expect  # type: ignore
+
+                self._override_test(
+                    self,
+                    dummy_tester,
+                    golden,
+                )
+            else:
+                golden_stdout, golden_output = self._eval(
+                    with_captured_stdout(golden)
+                )  # type: str, Output
+                if self._expect is not None:
+                    if self._override_check is not None:
+                        self._override_check(self, self._expect, golden_output)
+                    else:
+                        self.assertEqual(golden_output, self._expect)
+
+                if self._expect_stdout is not None:
+                    if isinstance(self._expect_stdout, str):
+                        self.assertEqual(golden_stdout, self._expect_stdout)
+                    elif isinstance(self._expect_stdout, Sequence):
+                        self.assertEqual(
+                            golden_stdout.splitlines(), list(self._expect_stdout)
+                        )
 
 
 class _TestInputGroup(Generic[Output]):
@@ -391,6 +480,7 @@ class Problem(Generic[Output]):
         aga_override_test: Optional[
             Callable[[TestCase, Callable[..., Output], Callable[..., Output]], None]
         ] = None,
+        aga_description: Optional[str] = None,
         **kwargs: Any,
     ) -> None:
         """Add a test case to the current group.
@@ -410,6 +500,7 @@ class Problem(Generic[Output]):
             aga_override_check=aga_override_check,
             aga_override_test=aga_override_test,
             aga_mock_input=self._config.problem.mock_input,
+            aga_description=aga_description,
             **kwargs,
         )
         self._ungrouped_tests.append(case)
@@ -591,10 +682,7 @@ def _check_reserved_keyword(kwd: str) -> None:
         )
 
 
-def test_case(
-    *args: Any,
-    **kwargs: Any,
-) -> Callable[[Problem[Output]], Problem[Output]]:
+class _TestCase:
     r"""Declare a specific test case for some problem.
 
     Parameters
@@ -619,6 +707,9 @@ def test_case(
     aga_name : Optional[str]
         The test case's name. If `None`, defaults to "Test {inputs}", where {inputs} is
         a comma-separated list of args and kwargs.
+    aga_description: Optional[str]
+        The detailed description for the test case. It will be displayed under the
+        test case name and thus supports longer descriptions.
     aga_weight : int
         The test case's relative weight to the group's score. See :ref:`Determining
         Score` for details.
@@ -642,22 +733,143 @@ def test_case(
     Callable[[Problem[T]], Problem[T]]
         A decorator which adds the test case to a problem.
     """
-    for kwd, _ in kwargs.items():
-        _check_reserved_keyword(kwd)
 
-    def outer(prob: Problem[Output]) -> Problem[Output]:
+    def __init__(self, *args: Any, **kwargs: Any):
+        """Initialize a Param."""
+        self._args = args
+        self._kwargs = kwargs
+
+    @property
+    def args(self) -> Tuple[Any, ...]:
+        """Return the arguments to be passed to the functions under test."""
+        return self._args
+
+    @property
+    def kwargs(self) -> Dict[str, Any]:
+        """Return the keyword arguments to be passed to the functions under test."""
+        return self._kwargs
+
+    def update_kwargs(self, **kwargs: Any) -> _TestCase:
+        """Update the keyword arguments to be passed to the functions under test."""
+        self._kwargs.update(kwargs)
+        return self
+
+    def generate_test_case(self, prob: Problem[Output]) -> Problem[Output]:
+        """Generate a test case for the given problem."""
+        return self(prob)
+
+    def __call__(self, prob: Problem[Output]) -> Problem[Output]:
+        """Add the test case to the given problem."""
+        for kwd, _ in self.kwargs.items():
+            _check_reserved_keyword(kwd)
+
         prob.add_test_case(
-            *args,
-            **kwargs,
+            *self.args,
+            **self.kwargs,
         )
+
         return prob
 
-    return outer
+
+param = _TestCase  # pylint: disable=invalid-name
+test_case = _TestCase  # pylint: disable=invalid-name
+
+
+def _parse_params(*args: Iterable[Any], **kwargs: Any) -> List[_TestCase]:
+    """Parse the parameters for param sequence."""
+    if kwargs:
+        raise ValueError("aga_params=True ignores non-aga kwargs")
+
+    if len(args) != 1:
+        raise ValueError(
+            "aga_params=True requires exactly one iterable of sets of parameters"
+        )
+
+    return list(arg if isinstance(arg, _TestCase) else param(*arg) for arg in args[0])
+
+
+def _parse_zip_or_product(
+    *args: Iterable[Any],
+    aga_product: bool = False,
+    aga_zip: bool = False,
+    **kwargs: Any,
+) -> List[_TestCase]:
+    """Parse parameters for zip or product."""
+    if not aga_zip ^ aga_product:
+        raise ValueError("exactly one of aga_zip or aga_product must be True")
+
+    combinator = product if aga_product else zip
+
+    # ok so if the combinator is product
+    # we are taking the cartesian product for all args and kwargs
+    # and if the combinator is zip,
+    # we are zipping all the args and kwargs, if there are any
+    combined_args = list(combinator(*args))
+
+    combined_kwargs = list(combinator(*kwargs.values()))
+
+    # ======= validation checks =======
+    if combinator is zip:
+        # create empty args for zip if there are no args
+        if combined_args and combined_kwargs:
+            if len(combined_args) != len(combined_kwargs):
+                raise ValueError('length of "args" and "kwargs" must match in zip mode')
+        elif combined_args:
+            combined_kwargs = [()] * len(combined_args)
+        elif combined_kwargs:
+            combined_args = [()] * len(combined_kwargs)
+
+    all_args_and_kwargs = list(combinator(combined_args, combined_kwargs))
+
+    # ======= zipping all the args together =======
+    return list(
+        param(*curr_args, **dict(zip(kwargs.keys(), curr_kwargs)))
+        for (curr_args, curr_kwargs) in all_args_and_kwargs
+    )
+
+
+def _add_aga_kwargs(aga_kwargs: Dict[str, Any], final_params: List[_TestCase]) -> None:
+    """Add aga_kwargs to the finalized parameters."""
+    # process aga input type
+    for aga_kwarg_key, aga_kwarg_value in aga_kwargs.items():
+        if isinstance(aga_kwarg_value, Iterable) and not isinstance(
+            aga_kwarg_value, str
+        ):
+            aga_kwargs[aga_kwarg_key] = list(aga_kwarg_value)
+        else:
+            aga_kwargs[aga_kwarg_key] = [aga_kwarg_value] * len(final_params)
+
+    # validate aga input type
+    if not all(
+        len(aga_kwarg_value) == len(final_params)
+        for aga_kwarg_value in aga_kwargs.values()
+    ):
+        # the length of the kwargs should be equal to the number of test cases
+        # i.e. the length of the combined args
+        raise ValueError(
+            f"all aga_ keyword args must have the same length as the test cases, "
+            f"which is {len(final_params)}"
+        )
+
+    # the aga kwargs list we want
+    aga_kwargs_list: List[Dict[str, Any]] = [
+        dict(zip(aga_kwargs.keys(), aga_kwarg_value))
+        for aga_kwarg_value in zip(*aga_kwargs.values())
+    ]
+
+    if not aga_kwargs_list:
+        # generate default aga kwargs dict if there are no aga kwargs
+        aga_kwargs_list = [{} for _ in final_params]
+
+    for final_param, kwargs in zip(final_params, aga_kwargs_list):
+        final_param.update_kwargs(**kwargs)
 
 
 def test_cases(
     *args: Iterable[Any],
-    aga_product: bool = True,
+    aga_product: bool = False,
+    aga_zip: bool = False,
+    aga_params: bool = False,
     **kwargs: Any,
 ) -> Callable[[Problem[Output]], Problem[Output]]:
     r"""Generate many test cases programmatically, from generators of inputs.
@@ -669,7 +881,12 @@ def test_cases(
     aga_product : bool
         Whether to take a cartesian product of the generators (creating one test case
         for each set of inputs in the product), or to zip them (iterate through each
-        generator in sequence). Default `True`.
+        generator in sequence). Default `False`.
+    aga_zip : bool
+        Whether to zip the input generators. Default `False`.
+    aga_params : bool
+        Whether to treat the input generators as a single iterable of sets of
+        parameters. Default `False`.
     kwargs :
         `aga_` keywords have their meaning inherited from `test_case`, and are
         applied to each test case generated by this function. Singleton value and
@@ -684,90 +901,52 @@ def test_cases(
     Callable[[Problem[T]], Problem[T]]
         A decorator which adds the test cases to a problem.
     """
+    if not (aga_product ^ aga_zip ^ aga_params) or (
+        aga_product and aga_zip and aga_params
+    ):
+        raise ValueError(
+            "Exactly one of aga_product, aga_zip, or aga_params must be True. \n"
+            f"You got: "
+            f"aga_product={aga_product}, aga_zip={aga_zip}, aga_params={aga_params}"
+        )
+
+    # pop aga keywords out
+    aga_kwargs_dict = {
+        kwd: kwargs.pop(kwd) for kwd in AGA_RESERVED_KEYWORDS if kwd in kwargs
+    }
+
+    final_params: List[_TestCase]
+
+    if aga_params:
+        # build final params
+        # So we're allowing [param(1, 2), [3, 4]] as a valid input
+        final_params = _parse_params(*args, **kwargs)
+    else:
+        final_params = _parse_zip_or_product(
+            *args, aga_zip=aga_zip, aga_product=aga_product, **kwargs
+        )
+
+    _add_aga_kwargs(aga_kwargs_dict, final_params)
 
     def outer(prob: Problem[Output]) -> Problem[Output]:
-        # ok so if the combinator is product
-        # we are taking the cartesian product for all args and kwargs
-        # and if the combinator is zip,
-        # we are zipping all the args and kwargs, if there are any
-        combinator = product if aga_product else zip
-
-        combined_args = list(combinator(*args))
-
-        # pop aga keywords out
-        aga_kwargs_dict = {
-            kwd: kwargs.pop(kwd) for kwd in AGA_RESERVED_KEYWORDS if kwd in kwargs
-        }
-
-        combined_kwargs = list(combinator(*kwargs.values()))
-
-        # ======= validation checks =======
-        if combinator is zip:
-            # create empty args for zip if there are no args
-            if combined_args and combined_kwargs:
-                if len(combined_args) != len(combined_kwargs):
-                    raise ValueError(
-                        'length of "args" and "kwargs" must match in zip mode'
-                    )
-            elif combined_args:
-                combined_kwargs = [()] * len(combined_args)
-            elif combined_kwargs:
-                combined_args = [()] * len(combined_kwargs)
-
-        all_args_and_kwargs = list(combinator(combined_args, combined_kwargs))
-
-        # process aga input type
-        for aga_kwarg_key, aga_kwarg_value in aga_kwargs_dict.items():
-            if isinstance(aga_kwarg_value, Iterable) and not isinstance(
-                aga_kwarg_value, str
-            ):
-                aga_kwargs_dict[aga_kwarg_key] = list(aga_kwarg_value)
-            else:
-                aga_kwargs_dict[aga_kwarg_key] = [aga_kwarg_value] * len(
-                    all_args_and_kwargs
-                )
-
-        # validate aga input type
-        if not all(
-            len(aga_kwarg_value) == len(all_args_and_kwargs)
-            for aga_kwarg_value in aga_kwargs_dict.values()
-        ):
-            # the length of the kwargs should be equal to the number of test cases
-            # i.e. the length of the combined args
-            raise ValueError(
-                f"all aga_ keyword args must have the same length as the test cases, "
-                f"which is {len(combined_args)}"
-            )
-
-        # the aga kwargs list we want
-        aga_kwargs_list: List[Dict[str, Any]] = [
-            dict(zip(aga_kwargs_dict.keys(), aga_kwarg_value))
-            for aga_kwarg_value in zip(*aga_kwargs_dict.values())
-        ]
-
-        if not aga_kwargs_list:
-            # generate default aga kwargs dict if there are no aga kwargs
-            aga_kwargs_list = [{} for _ in all_args_and_kwargs]
-
-        # ======= zipping all the args together =======
-        total_args = list(zip(all_args_and_kwargs, aga_kwargs_list))
-
-        for (curr_args, curr_kwargs), aga_kwargs in total_args:
-            # we have to reassemble a kwargs dictionary that actually has keys
-            final_kwargs = dict(zip(kwargs.keys(), curr_kwargs), **aga_kwargs)
-
-            # i think we have to enumerate all of the arguments to test_case by
-            # hand; we can't capture them separately from the test case kwargs
-            prob = test_case(
-                *curr_args,
-                **final_kwargs,
-            )(prob)
+        """Add the test cases to the problem."""
+        for final_param in final_params:
+            prob = final_param.generate_test_case(prob)
 
         return prob
 
     return outer
 
 
+test_cases_params = partial(test_cases, aga_params=True)
+test_cases_zip = partial(test_cases, aga_zip=True)
+test_cases_product = partial(test_cases, aga_product=True)
+
+test_cases.params = test_cases_params  # type: ignore
+test_cases.product = test_cases_product  # type: ignore
+test_cases.zip = test_cases_zip  # type: ignore
+
+
 def group(
     weight: int = 1,
     value: float = 0.0,
@@ -780,8 +959,10 @@ def group(
     weight : int
         The group's relative weight to the problem's score. See :ref:`Determining Score`
         for details.
-    value : int
+    value : float
         The group's absolute score. See :ref:`Determining Score` for details.
+    extra_credit : float
+        The group's extra credit. See :ref:`Determining Score` for details.
 
     Returns
     -------
